# Generated by Django 5.2.5 on 2025-08-22 12:32
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0016_tags_and_changelog_for_snapshots"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsync",
            name="update_custom_fields",
            field=models.BooleanField(default=True),
        ),
    ]
